#define TI89_VERSION 0

#define RANDOM_STRENGTH 4
#define VTI_TIMING 0

#if TI89_VERSION == 1
#define USE_TI89
#define calc(x, y) (x)
#else
#define USE_TI92PLUS
#define USE_V200
#define calc(x, y) (y)
#endif

#include <tigcclib.h>

#include "extgraph.h"
#include "extrand.h"

#define VERSION_MAJOR 2
#define VERSION_MINOR 1
#define DATA_NAME SYMSTR("skidat")


typedef struct
{
	char initials[3];
	unsigned char calc : 1, random_strength : 4, padding : 3;
} FILE_VERSION;

typedef struct
{
	struct
	{
		short size;
		FILE_VERSION version;
	} header;
	struct
	{
		unsigned short scores[4];
		short width;
		RANDOM_BUFFER buffer;
	} data;
	struct
	{
		char hash[17];
		char type[7];
	} footer;
} DATA_FILE;

static short MainMenu(short[], short *);
static short Wrap(short, short, short);
static unsigned long PlayGame(char);
static void Pause(void);
void SleepMS_HW1(unsigned short);
void SleepMS_HW2(unsigned short);

void _main(void)
{
	char s_buffer[18];
	char *lcd;
	short old_font;
	short ok = 0;
	unsigned long score;
	DATA_FILE *filep = NULL;
	FILE_VERSION Check = {{'P', 'J', 'R'}, TI89_VERSION, RANDOM_STRENGTH, 0};
	MD5_CTX ctx;
	SYM_ENTRY *file_vat;
	INT_HANDLER OldInt1, OldInt5;

	lcd = malloc(sizeof(LCD_BUFFER));
	if(lcd == NULL)
	{
		ST_helpMsg("ERROR: memory low");
		return;
	}
	FastCopyScreen(LCD_MEM, lcd);

	file_vat = SymFindPtr(DATA_NAME, 0); /*see if the high score file exists*/
	if(file_vat != NULL) /*if it does*/
	{
		if(file_vat->flags.bits.archived) /*and it is archived*/
		{
			EM_moveSymFromExtMem(DATA_NAME, HS_NULL); /*unarchive it*/
		}
		filep = HeapDeref(file_vat->handle);

		MD5Init(&ctx);
		MD5Update(&ctx, (char *)&filep->data, sizeof(filep->data));
		MD5Final(s_buffer, &ctx);
	}

	if(file_vat == NULL || memcmp(&filep->header.version, &Check, sizeof(FILE_VERSION)) ||
	memcmp(filep->footer.hash, s_buffer, 16)) /*if the file does not exist, is the wrong type, or has been
	edited*/
	{
		RANDOM_BUFFER buffer;

		randomize_long();

		SaveRandState(&buffer);
		/*make the file*/
		SymDel(DATA_NAME);
		file_vat = DerefSym(SymAdd(DATA_NAME));
		file_vat->handle = HeapAlloc(sizeof(DATA_FILE));
		if(file_vat->handle == H_NULL)
		{
			ST_helpMsg("ERROR: memory low");
			free(lcd);
			return;
		}
		filep = HeapDeref(file_vat->handle);
		memset(filep, 0, sizeof(DATA_FILE));

		filep->header.size = sizeof(DATA_FILE) - 2;
		memcpy(&filep->header.version, &Check, sizeof(FILE_VERSION));

		filep->data.width = 3;
		memcpy(&filep->data.buffer, &buffer, sizeof(RANDOM_BUFFER));

		memcpy(filep->footer.type, &(char[]){0, 'D', 'a', 't', 'a', 0, OTH_TAG}, 7);
	}

	LoadRandState(&filep->data.buffer);

	old_font = FontSetSys(F_4x6); /*set the font to 4x6*/
	OldInt1 = GetIntVec(AUTO_INT(1));
	OldInt5 = GetIntVec(AUTO_INT(5));
	SetIntVec(AUTO_INT(1), DUMMY_HANDLER); /*turn off the status line indicators by
	turning off auto-int 1*/
	SetIntVec(AUTO_INT(5), DUMMY_HANDLER); /*improve performance of rowread by turning off
	auto-int 5*/

	do
	{
		ok = MainMenu(filep->data.scores, &filep->data.width);
		if(ok == 1) /*if [Enter] was pressed*/
		{
			ClrScr();
			score = PlayGame(filep->data.width + 5); /*call the actual game and return the
			score*/
			ok = 0; /*clear out the [2nd] press*/

			ClrScr();
			sprintf(s_buffer, "Score: %lu", score);
			DrawStr((calc(160, 240) - DrawStrWidth(s_buffer, F_4x6)) / 2, 0, s_buffer,
			A_NORMAL);
			if(score > filep->data.scores[filep->data.width])
			{
				DrawStr(calc(42, 82), 6, "You have a high score!", A_NORMAL);
				filep->data.scores[filep->data.width] = score;
			}
			DrawStr(calc(40, 80), 12, "Press [2nd] to continue.", A_NORMAL);
			Pause();
		}
	} while(ok != 2); /*loop while [ESC] not pressed*/

	SaveRandState(&filep->data.buffer);

	MD5Init(&ctx);
	MD5Update(&ctx, (char *)&filep->data, sizeof(filep->data));
	MD5Final(filep->footer.hash, &ctx);

	EM_moveSymToExtMem(DATA_NAME, HS_NULL); /*archive the file*/

	SetIntVec(AUTO_INT(1), OldInt1); /*turn the status line stuff back on*/
	SetIntVec(AUTO_INT(5), OldInt5); /*turn the timer back on*/
	FontSetSys(old_font); /*set the font back to what it was*/
	FastCopyScreen(lcd, LCD_MEM);
	ST_helpMsg(EXTGRAPH_VERSION_PWDSTR);
	free(lcd);
}

short MainMenu(short high_score[], short *widthp)
{
	char buffer[9];
	short i;
	short key;
	short ok = 0;
	short row = 3 - *widthp;
	static const char *difficulty[4] = {"(Easy)", "(Medium)", "(Hard)", "(Impossible)"};
	static short row_ends[4] = {48, 59, 49, 68};

	ClrScr();
	sprintf(buffer, "Ski v%d.%d", VERSION_MAJOR, VERSION_MINOR);
	DrawStr((calc(160, 240) - DrawStrWidth(buffer, F_4x6)) / 2, 0, buffer, A_NORMAL);
	DrawStr(calc(46, 86), 6, "By Peter J. Rowe", A_NORMAL);
	DrawStr(calc(49, 89), 12, "mig53@yahoo.com", A_NORMAL);
	DrawStr(calc(61, 101), 24, "High Scores", A_NORMAL);
	DrawStr(calc(33, 73), 82, "Use \x15/\x16 to control the skier.", A_NORMAL);
	DrawStr(calc(33, 73), 88, "Use \x17/\x18 to select difficulty.", A_NORMAL);
	DrawStr(calc(19, 59), 94, "Press [2nd] to start or [ESC] to quit.", A_NORMAL);
	for(i = 0; i < 4; i++) /*print out the high scores*/
	{
		printf_xy(0, 30 + 12 * i, "Width %d %s", 8 - i,
		difficulty[i]);
		printf_xy(0, 36 + 12 * i, "%u", high_score[3 - i]);
	}

	FastFilledRect_Invert_R(LCD_MEM, 0, 12 * row + 29, row_ends[row], 12 * row + 35);

	while(!ok) /*do while 2nd or esc not pressed*/
	{
		key = _rowread(0xFFFE);

		if(key & calc(5, 160))
		{
			FastFilledRect_Invert_R(LCD_MEM, 0, 12 * row + 29, row_ends[row], 12 * row + 35);
		}
		if(key & calc(1, 32)) /*up*/
		{
			row--;
			while(_rowread(0xFFFE) & calc(1, 32));
		}
		if(key & calc(4, 128)) /*down*/
		{
			row++;
			while(_rowread(0xFFFE) & calc(4, 128));
		}
		row = Wrap(row, 0, 3);
		if(key & calc(5, 160))
		{
			FastFilledRect_Invert_R(LCD_MEM, 0, 12 * row + 29, row_ends[row], 12 * row + 35);
		}
		if(key & calc(16, 1))
		{
			ok = 1;
			*widthp = 3 - row;
			while(_rowread(0xFFFE) & calc(16, 1));
		}
		if(_rowread(calc(0xFFBF, 0xFEFF)) & calc(1, 64)) /*esc*/
		{
			ok = 2;
			*widthp = 3 - row;
			while(_rowread(calc(0xFFBF, 0xFEFF)) & calc(1, 64));
		}
	}

	return(ok);
}

short Wrap(short input, short lower, short upper) /*take the input value and put it
between the lower and upper bound with wrapping around the ends*/
{
	return(input >= lower ? (input <= upper ? input : lower) : upper);
}

unsigned long PlayGame(char width)
{
	short key;
	short skier = calc(77, 117);
	unsigned char walls[calc(20, 25)];
	unsigned long score = 0;
	static char random_move[5] = {-1, -1, 0, 1, 1};
	static unsigned char ski_icon[5] = {0x20, 0x50, 0x70, 0x50, 0x20};
	static unsigned char wall_icon[5] = {0x40, 0x40, 0x40, 0x40, 0x40};

	memset(walls, calc(76, 116), calc(20, 25));
	FastLine_Draw_R(LCD_MEM, calc(76, 116), 0, calc(76, 116), calc(99, 127));
	FastLine_Draw_R(LCD_MEM, calc(76 + width, 116 + width), 0, calc(76 + width, 116 + width),
	calc(99, 127));

	do
	{
		memmove(LCD_MEM, LCD_MEM + 5 * 30, sizeof(LCD_BUFFER) - 5 * 30);
		memmove(&walls[0], &walls[1], calc(19, 24));

		walls[calc(19, 24)] += random_move[random_long(5)]; /*randomly move the bottom
		wall*/
		walls[calc(19, 24)] = max(min(walls[calc(19, 24)], calc(159, 239) - width), 1);
		/*make sure the bottom wall stays within the screen*/
		/*draw the last wall*/
		Sprite8_OR_R(walls[calc(19, 24)] - 1, calc(95, 120), 5, wall_icon, LCD_MEM);
		Sprite8_OR_R(walls[calc(19, 24)] + width - 1, calc(95, 120), 5, wall_icon, LCD_MEM);

		key = _rowread(0xFFFE); /*arrow keys*/

		if(key&calc(2, 16))
		{
			skier--; /*if left pressed*/
		}
		if(key&calc(8, 64))
		{
			skier++; /*if right pressed*/
		}

		skier = max(min(skier, calc(157, 237)), 1); /*keep the skier on the screeen*/
		Sprite8_OR_R(skier - 1, 0, 5, ski_icon, LCD_MEM); /*draw the skier*/
		score++; /*increase the score*/

		#if VTI_TIMING == 1
			SleepMS_HW1(150);
		#else
			HW_VERSION == 1 ? SleepMS_HW1(36) : SleepMS_HW2(36);
		#endif
	} while((skier > walls[0]) && (skier < walls[0] + width - 2));
	/*loop while the skier is within the walls*/

	return(score);
}

void Pause(void)
{
	while(_rowread(0xFFFE) & calc(16, 1)); /*wait until 2nd not pressed*/
	while(!(_rowread(0xFFFE) & calc(16, 1))); /*wait until 2nd pressed*/
	while(_rowread(0xFFFE) & calc(16, 1)); /*wait until 2nd not pressed*/
}